

%**************************************************************************
%**************************************************************************
%Calculos que se realizan una sola vez.
%**************************************************************************
%%*************************************************************************

kp = 1;      %Constante de aproximacion proporcional
global Af;

robots_avanzando = zeros(1,nRobots);


%************************************
%Obtencion de la matriz de inercia J.
%************************************
J=ObtenerMatrizInerciaJ(dim, mu, I0);


%*****************************************
%Obtencion del spline en estado de reposo.
%*****************************************
%-Creacion de la entrada del algoritmo de interpolacin a partir de los
% puntos de la manguera y las posiciones de los robots, todo ello en
% en estado de reposo.
%-Formato de los puntos de interpolacion con 3 robots y 2 segmentos, ej.:
%         [{robot 1};{segmento 1};{robot2};{segmento2};{robot3}]
puntosInterpolacion = ObtenerPuntosInterpolacion (nSegmentos,... 
                                                  robots_reposo,...
                                                  segmentos_reposo);

%Interpolacion de los puntos del manguera para obtener los puntos de
%control de la manguera en estado de reposo.
bsp_reposo = creacionBspline(1, puntosInterpolacion);


%******************************************
%Obtencion del spline en estado de inicial.
%******************************************
%-Creacion de la entrada del algoritmo de interpolacin a partir de los
% puntos de la manguera y las posiciones de los robots, todo ello en
% en estado inicial.
%-Formato de los puntos de interpolacion con 3 robots y 2 segmentos, ej.:
%         [{robot 1};{segmento 1};{robot2};{segmento2};{robot3}]
puntosInterpolacion = ObtenerPuntosInterpolacion (nSegmentos,... 
                                                  robots_inicial,...
                                                  segmentos_inicial);

%Interpolacion de los puntos del manguera para obtener los puntos de
%control de la manguera en la posicion inicial.
bsp_inicial = creacionBspline(1, puntosInterpolacion);


%Obtencion en variables de rapido acceso de los campos mas relevantes 
%del bsp_inicial.
nudos_inicial = bsp_inicial.nudos;

n_NumPtosControl = size(bsp_inicial.ptosControl,1);
m_NumNudos = size(nudos_inicial);


%**************************************************************************
%Calculo de la matriz Jacobiana.
%**************************************************************************
[Jir, Jri] = ObtenerMatrizJacobiana(n_NumPtosControl, nRobots, ...
                                    u_robots, p, nudos_inicial, L, dim);


%**************************************************************************
%Creacion de vectores columna de algunas magnitudes.
%**************************************************************************
%Con puntos de control: (ptoControl_1; ptoControl_2;... pto_Control_n)
%  coordenadas de cada punto en {x; y; z; theta}

%Posiciones de los puntos de control del spline en reposo.
q_reposo = ObtenerVectorColDeMatrizNx2plus(bsp_reposo.ptosControl, dim);

%Posiciones de los puntos de control del spline en la posicion inicial.
q_inicial = ObtenerVectorColDeMatrizNx2plus(bsp_inicial.ptosControl, dim);

%Posiciones iniciales de los robots.
pos_robots_inicial = ...
                 ObtenerVectorColDeMatrizNx2plus(robots_inicial(:,1:2),1);


%**************************************************************************
%Obtencion de las velocidades vectoriales iniciales de los robots.
%**************************************************************************
v_robots_inicial = [];
for i=1:nRobots
    v_robots_inicial = [v_robots_inicial; 
                        norm_v_robots_inicial(i)*ori_robots_inicial(:,i)];
end


%**************************************************************************
%Calculo de los polinomios generadores (funciones base) y sus derivadas.
%**************************************************************************
%Calculo de los polinomios generadores (funciones base).
Niu = ObtenerNiuOrden(L, Discret, n_NumPtosControl, p, nudos_inicial, 0);

%Calculo de las derivadas primeras de los polinomios generadores en u.
Niu_prima = ObtenerNiuOrden(L, Discret, n_NumPtosControl, p, ...
                            nudos_inicial, 1);

%Calculo de las derivadas segundas de los polinomios generadores en u.
Niu_primaprima = ObtenerNiuOrden(L, Discret, n_NumPtosControl, p, ...
                                 nudos_inicial, 2);

                             
%**************************************************************************
%Calculo de la derivada del spline en u en reposo.
%**************************************************************************
RprimaReposo = ObtenerRprima (dim, n_NumPtosControl, q_reposo, ...
                              Niu_prima, L, DeltaU);
norm_RprimaReposo = ObtenerNormaPorCols(RprimaReposo);

                                                      
%**************************************************************************
%Calculo de la "matriz de masas".
%**************************************************************************
M=get_M_JM(n_NumPtosControl, J, DeltaU, L, Niu);
    

%**************************************************************************
%Inicializacion de variables para contener los resultados en cada ciclo.
%**************************************************************************
robots=robots_inicial; 
pos_robots=pos_robots_inicial;
ori_robots=ori_robots_inicial;
norm_v_robots=norm_v_robots_inicial;
v_robots = v_robots_inicial;
q=q_inicial;
v=zeros(n_NumPtosControl*dim,1);
v(1:2)=norm_v_robots_inicial(1)*ori_robots_inicial(1:2,1); %Primer punto de 
                                                %control acompana al lider.

%**************************************************************************
%Inicializacion de matrices de resultados.
%**************************************************************************
MatricesResultados_inicializar;


%**************************************************************************
%**************************************************************************
% Ciclos de ejecucion.
%**************************************************************************
%**************************************************************************
for h = 1:Inf
    tic   

    %**********************************************************************
    %Gestion del aumento del tamano de las matrices.
    %**********************************************************************
    %Aumentan la primera vez que entra en el ciclo de ejecucion y cada 
    %'frec_aumento_matrices_resultados' ciclos, aumentando en 
    %'aumento_matrices_resultados' columnas.
    if h==1 || mod(h, frec_aumento_matrices_resultados)==0
        MatricesResultados_incrementar;
    end
    
   
    %**********************************************************************
    %Almacenamiento de magnitudes del ciclo de ejecucion anterior.
    %**********************************************************************
    %Velocidad vectorial de los robots.
    vant_robots = v_robots;
        
    
    %**********************************************************************
    %**********************************************************************
    % INICIO DEL BLOQUE DE CONTROL DE LOS ROBOTS SEGUDORES.
    %%*********************************************************************
    %%*********************************************************************
    
    %Calculo de la distancia lineal a la que estan cada robot del anterior
    %(se calcula para todos los robots menos para el lider).
    d_robots = zeros(nRobots-1,1);
    for i=1:nRobots-1
        d_robots(i) = distancia2d(robots(i,:),robots(i+1,:));            
    end
       
    
    %Calculo del modulo de la velocidad deseada para cada robot menos
    %el lider (norm_vd_robots). Para el lider, viene dado como parametro.
    if (h==1 || mod(h, frec_control_seguidores)==0)    
        d_deseada_aux=(L/nSegmentos)*...
                      (porcentaje_distancia_a_robot_previo/100);
        d_robots_deseadas = d_deseada_aux*ones(nRobots-1,1);
        
        %La norma de la velocidad del primer robot viene como parametro.
        norm_v_robots(2:nRobots) = ...
                   (ones(nRobots-1,1)-d_robots_deseadas + d_robots).* ...
                     norm_v_robots(1:nRobots-1);
    end
    %En su entrada i-esima (considerar 'frec_control') asigna velocidad
    %al robot (i+1) poniendo para el un valor distinto de cero en la
    %matriz norm_v_robots, es decir,...
    %(1a entrada arranca robot2, 2a entrada arranca robot3, etc...).
    
    
    %Calculo de la distancia vectorial de cada robot a su predecesor
    %(de todos los robots menos el lider), y se normaliza para obtener 
    %sus vectores directores.
    v_rant_ract_norm = robots(1:nRobots-1,:)-robots(2:nRobots,:);
    v_rant_ract_norm = normalizarFilas(v_rant_ract_norm);   
    
    %Obtencion de las orientaciones deseadas de los robots que siguen
    %al lider (la del lider no se determina aqui). A la orientacion
    %actual se aade el error existente entre la orientacion actual y la
    %del predecersor, con una constante de proporcionalidad 'kp'.
    ori_robots(1:2,2:nRobots) = ori_robots(1:2,2:nRobots)+...
                         kp*(v_rant_ract_norm'-ori_robots(1:2,2:nRobots));
                     
    
    %Calculo de la velocidad vectorial de cada robot:
    %-Al LIDER, en cada ciclo se le asigna la norma por su orientacion.
    %-Al RESTO, solo se mira cuando ponerlos en marcha, y se hace una 
    %           unica vez por robot:
    %   deben estar parados y su distancia con el predeceror mayor que
    %   una dada, en ese caso se fija a su norma por su orientacion.
    for i=1:nRobots
        indice = dim*(i-1);
        if ((i>1) && (robots_avanzando(i)==0))
            d_deseada_aux=(L/nSegmentos)*...
                           (porcentaje_distancia_a_robot_previo/100);
            if (d_robots(i-1)>d_deseada_aux)
                v_robots(indice+1:indice+dim,1)=norm_v_robots(i)*...
                                                ori_robots(:,i);
                robots_avanzando(i)=1;
            end
        else
            v_robots(indice+1:indice+dim,1)=norm_v_robots(i)*...
                                            ori_robots(:,i);
        end
    end

    %**********************************************************************
    % FIN DEL BLOQUE DE CONTROL DE LOS ROBOTS SEGUIDORES.
    %**********************************************************************
    
    
    %**********************************************************************
    %**********************************************************************
    %Postprocesado de la velocidad determinada por el bloque de control.
    %**********************************************************************
    %**********************************************************************
    
    %**********************************************************************
    %Limitacion de la velocidad de todos los robots en todas sus 
    %componentes de modo solidario.
    %**********************************************************************
    v_robots = ObtenerVelocidadesAcotadas(v_robots, Vmaxima);
    
    
    %**********************************************************************
    %Obtencion de las aceleraciones mediante la derivacion de las 
    %velocidades de los robots.
    %**********************************************************************
    a_robots=(v_robots-vant_robots)/DeltaT;

    
    %**********************************************************************
    %Limitacion de la aceleracion de todos los robots, solidariamente 
    %robot a robot en todas sus componentes, para que cada uno alcance 
    %la velocidad debida en el menor tiempo posible dentro del limite.
    %**********************************************************************
    a_robots = ObtenerAceleracionesAcotadas (dim, a_robots, Amaxima);
      
   
    %**********************************************************************
    %Recalculo de la velocidad de los robots: es la velocidad del ciclo
    %de simulacion anterior mas el incremento dado por la aceleracion 
    %acotada calculada en este ciclo.
    %**********************************************************************
    v_robots = vant_robots + a_robots*DeltaT;
    

    %**********************************************************************
    %Calculo de la aceleracion producida en los puntos de control por los
    %robots, una vez que se sabe la que tienen los robots.
    %**********************************************************************
    A = Jir*a_robots;

    
    %**********************************************************************
    %**********************************************************************
    %Calculo de las fuerzas derivadas de las energias potenciales 
    %producidas en la manguera.
    %**********************************************************************
    %**********************************************************************
        
    %**********************************************************************
    %Calculo de las derivadas de Q y R en funcion del parametro u.
    %**********************************************************************
    Rprima=ObtenerRprima(dim, n_NumPtosControl, q, Niu_prima, L, DeltaU);
    
    norm_Rprima = ObtenerNormaPorCols(Rprima);
    
    Qprimaprima=CalcularQprimaprima(dim, n_NumPtosControl, q,...
                                    Niu_primaprima, L, DeltaU);
     
                                
    %**********************************************************************
    %Calculo de las fuerzas inducidas por la energia potencial.
    % Ps (streching), fuerza debida al estiramiento,
    % Pt (twisting), fuerza debida a la torsion (NO CALCULADA),
    % Pb (bending), fuerza debida al curvamiento.
    %**********************************************************************
    Ps=get_Ps(dim, n_NumPtosControl, Rprima, Niu_prima, DeltaU, L, ES,...
              D, pi,norm_RprimaReposo,norm_Rprima,cs);
    Pb=get_Pb(dim, n_NumPtosControl, Qprimaprima, Rprima, Niu_prima,...
              Niu_primaprima, DeltaU, L, EB, D, pi,norm_Rprima);
    P=Pb+Ps;
    
    
    %**********************************************************************
    %**********************************************************************
    %DINAMICA INVERSA (CONTROL).
    %**********************************************************************
    %**********************************************************************
    
    %Obtencion de las fuerzas netas que deben resultar en los puntos de
    %control para alcanzar las aceleraciones deseadas en los mismos, 
    %teniendo en cuenta las fuerzas causadas por la energia potencial
    %de la manguera (P).
    F=M*A-P;
        
    %Obtencion de la deceleracion que se produce por rozamiento de la
    %manguera.
    Arozamiento = ObtenerAceleracionesPorRozamiento ...
                        (dim, n_NumPtosControl, v, coef_rozamiento);
    
                    
    %Obtencion de las fuerzas que se generan por efecto del rozamiento
    %de la manguera.
    Frozamiento = M*Arozamiento;
    
    %A la fuerza necesaria para alcanzar la aceleracion (F) antes 
    %calculada se le anade el efecto de la f. de rozamiento (Frozamiento).
    F = F-Frozamiento;
    
    %Calculo de las fuerzas que deben aplicar los robots sobre la manguera
    %a partir de las fuerzas generalizadas en los puntos de control.
    Fr = ObtenerFuerzasAcotadas (Jri, F, Fmax_robots);
    
    
    %**********************************************************************
    %**********************************************************************
    %DINAMICA DIRECTA (SIMULACION).
    %**********************************************************************
    %**********************************************************************
    
    %Obtencion de las nuevas F generalizadas (resultantes en los puntos
    %de control) al aplicar los robots una fuerza Fr.
    Fgeneralizadas = Jir*Fr;
    
    %Obtencion de la fuerza neta en cada punto de control, teniendo en 
    %cuenta aparte de las fuerzas provinientes de los robots, las debidas
    %al rozamiento y a la dinaminca interna de la manguera.
    Fneta = Fgeneralizadas + Frozamiento + P;
  
    
    %**********************************************************************
    %Calculo de las aceleraciones que experimentarian los ptos. de control
    %a partir de la fuerza neta resultante.
    %**********************************************************************
    %Calculo de una semilla de Af: a priori, las aceleraciones son las
    %producidas por la Fneta.
    Af=M\(Fneta);
    
    %Restriccion: Las aceleraciones en los puntos de contacto de la
    %             manguera con robots son las de los robots.
    [Af, aux, exitflag] = fmincon(@difR2,Af,[],[],Jri,a_robots,[],[],[],...
                             opciones_fmincon);
                    
    if exitflag==0
        %fmincon no converge.
        Terminacion_experimento=1;
        Cadena_Terminacion_experimento=...
           ['Fin porque fmincon no converge, exitflag=' num2str(exitflag)];
    end
   
    
    %**********************************************************************
    %Calculo de las nuevas velocidades y posiciones de los puntos de 
    %control en base a su aceleracion.
    %**********************************************************************
    %Integracion de las aceleraciones de los puntos de control para
    %obtener sus velocidades.
    v = v + Af*DeltaT;
    
    %Integracion de las velocidades de los puntos de control para obtener
    %sus posiciones.
    q = q + v*DeltaT;

    
    %**********************************************************************
    %Calculo de las nuevas posiciones de los robots en base a sus
    %velocidades.
    %**********************************************************************
    %Integracion de las velocidades de los robots para obtener sus
    %posiciones.
    pos_robots = pos_robots + v_robots*DeltaT;
    
    %Empaquetado de informacion de modo vector unidimensional a modo
    %matriz 'dim'-dimensional:
    %Posicion de los robots en matriz (nRobots,dim).
    robots = ObtenerMatrizNxdimDeVectorCol(pos_robots, dim); 

    
    %**********************************************************************
    %**********************************************************************
    %Almacenamiento de los valores relevantes del ciclo de ejecucion
    %en las matrices de resultados.
    %**********************************************************************
    %**********************************************************************
    MatricesResultados_actualizar;
    
    
    %**********************************************************************
    %**********************************************************************
    %BLOQUE PARA EL CONTROL DEL LIDER.
    %**********************************************************************
    %**********************************************************************
    
    if (h==1 || mod(h, frec_control_lider)==0)
        %Consigna para el primer robot.
        if ((pos_robots(1)<2) && (pos_robots(2)<-2))
            ori_robots(:,1) = [0;-1]*abs(pos_robots(2)+3)+ [1;0]*abs(-pos_robots(2)-2);
        elseif (pos_robots(1)>2.5)
            ori_robots(:,1) = [1;0]*abs(pos_robots(1)-3.5)+ [0;1]*abs(pos_robots(1)-2.5);
        end
        
        ori_robots(:,1) = ori_robots(:,1)/norm(ori_robots(:,1));
    end
    %**********************************************************************
    %**********************************************************************   
    
    %**********************************************************************
    %**********************************************************************
    %Control del bucle, trazas, interrupciones, etc.
    %**********************************************************************
    %**********************************************************************

    %Impresion de 'h' cada 'frec_traza_pantalla_h' ciclos.
    if mod(h,frec_traza_pantalla_h)==0
        disp(strcat(datestr(now),'-> h terminado: ', num2str(h)));
    end
        
    if mod(h,frec_traza_pantalla_tictoc)==0
        disp([datestr(now), '->',num2str(toc), ' sg.']);
    end
    
    %Control del numero de iteraciones que se estan haciendo.
    if h==limite_iteraciones
        disp(strcat('Fin por ''limite_iteraciones''=',num2str(h)));
        break
    end
    
    %Atencion al teclado.
    AtenderTeclado;
    
    %Gestion de la terminacion del experimento.
    if Terminacion_experimento~=0
        break;
    end
    
end

%**************************************************************************
%Limpieza de las matrices de resultados.
%**************************************************************************
MatricesResultados_limpiar;
